"""
This script is an example of automatic binary recording.
It covers the main steps from the scenario creation to the analysis with a Reven server.
"""

import argparse
from time import sleep
from urllib.parse import urlparse

import reven2
import reven2.preview.project_manager as project_manager

from utils.project_manager import (
    display_failed_resources_actions_log,
    display_resources_actions_progress,
    is_replay_finished,
)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Fully automated scenario creation")

    parser.add_argument("--url", required=True, help="The URL of the Reven project manager")
    parser.add_argument("--snapshot", type=int, required=True, help="The id of the snapshot to use for the scenario")
    parser.add_argument("--live-snp", required=True, help="Name of the live snapshot to use")
    parser.add_argument("--name", default="automatic-binary-scenario", help="The name of the scenario to create")
    parser.add_argument("--description", default="", help="Description of the scenario")
    parser.add_argument("--local-binary", help="Local path of the binary to record")
    parser.add_argument("--guest-binary", help="Guest path of the binary to record")
    parser.add_argument("--args", default=None, nargs="+", help="Args to give to the binary")
    parser.add_argument(
        "--asm-stub",
        default=None,
        action="store_true",
        help="Record using the ASM stubs instead of autodetecting the execution of the binary",
    )
    parser.add_argument(
        "--timeout-start",
        type=int,
        help="The maximum number of seconds to wait between the readyness of the recorder and the start of the record",
    )
    parser.add_argument(
        "--timeout-record", type=int, help="The maximum number of seconds to wait when recording before stopping it"
    )
    parser.add_argument(
        "--resources", default=None, nargs="+", help="Specific resources you want to replay. By default replay all"
    )
    parser.add_argument(
        "--actions", default=None, nargs="+", help="Specific resources you want to replay. By default replay all"
    )
    parser.add_argument(
        "--clean-on-failure", dest="clean_on_failure", action="store_true", help="Clean created resources upon failure"
    )
    parser.add_argument(
        "--export", dest="export", action="store_true", help="Export and delete the scenario at the end of the script"
    )

    args = parser.parse_args()

    if not (bool(args.local_binary is None) ^ bool(args.guest_binary is None)):
        print("You should provide either --local-binary or --guest-binary and not both")
        exit(1)

    try:
        # connect to the Reven project manager server
        pm = project_manager.ProjectManager(args.url)

        # check that we are using a QEMU snapshot because automatic record is **not supported** for Vbox scenario
        if pm.get_snapshot(args.snapshot)["type"] != "QEMU":
            print("Automatic and multiple record is only available with a QEMU snapshot")
            exit(1)

        #####################
        # SCENARIO CREATION #
        #####################

        scenario = None
        binary_file = None
        # create a new scenario with a name and a snapshot (required) and a description (optional)
        # extra optional arguments can be specified like `cdrom_mount_point` or `max_icount`
        try:
            scenario = pm.create_scenario(
                args.name,
                args.snapshot,
                description=args.description,
            )
            print('Scenario "%s" created with id %d' % (scenario["name"], scenario["id"]))

            if args.local_binary:
                # upload the binary to auto-record and set it as one of the input files of the scenario
                binary_file = pm.upload_file(args.local_binary)
                scenario = pm.update_scenario(
                    scenario["id"],
                    input_files=[
                        binary_file["id"],
                    ],
                )

            print()

            ##########################
            # END SCENARIOS CREATION #
            ##########################

            ##########
            # RECORD #
            ##########

            # First, start the snapshot
            # a specific (optional) QEMU live snapshot can be started (if specified like here)
            # extra optional arguments can be specified, like `network`, `ram_size`, `custom_options`
            response = pm.start_qemu_snapshot_session(args.snapshot, live_snapshot=args.live_snp)
            print(response["message"])

            # we need to keep the VM session to start the auto-record and stop the session later
            session_id = response["session"]["id"]
            try:
                params = {
                    "qemu_session_id": session_id,
                }

                if args.timeout_start:
                    params["timeout_start"] = int(args.timeout_start)
                if args.timeout_record:
                    params["timeout_record"] = int(args.timeout_record)

                if binary_file is not None:
                    binary_name = binary_file["id"]
                else:
                    binary_name = args.guest_binary

                if args.asm_stub is None:
                    params["binary_name"] = binary_name

                # The autorun isn't mandatory
                # The start of the binary could be manual inside the VM
                # The binary to autorun could also be another binary, not necessarly the one we want to record
                # e.g a dropper unpacking a binary and launching it, the binary to record could be the one unpacked
                params["autorun_binary"] = binary_name

                params["autorun_files"] = list(map(lambda file: file["id"], scenario["input_files"]))

                if args.args is not None:
                    params["autorun_args"] = args.args

                auto_record_task = None
                try:
                    if args.asm_stub is None:
                        auto_record_task = pm.auto_record_binary(**params)["task"]
                    else:
                        auto_record_task = pm.auto_record_asm_stub(**params)["task"]

                    print("Record started")

                    while not auto_record_task["finished"]:
                        sleep(1)

                        # retrieve the task with updated information
                        auto_record_task = pm.get_task(auto_record_task["id"])
                        print("Recorder status: %s" % auto_record_task["display_recorder_status"])

                    try:
                        if auto_record_task["status"] in ["ABORTED", "FAILURE"]:
                            raise RuntimeError("Auto-record failed: %s" % auto_record_task["fail_reason"])

                        # Commit the record into the scenario
                        response = pm.commit_record(session_id, scenario["id"], record=auto_record_task["record_name"])
                        print(response["message"])

                        # When using the autorun, the files are copied into "C:\\reven" on Windows
                        scenario = pm.update_scenario(scenario["id"], cdrom_mount_point="C:\\reven")

                    finally:
                        auto_record_task = None

                finally:
                    if auto_record_task is not None:
                        response = pm.cancel_task(auto_record_task["id"])
                        print(response["message"])

            finally:
                # stop the session when we are done recording the scenario
                response = pm.stop_session(session_id)
                print(response["message"])

            print()

            ##############
            # END RECORD #
            ##############

            ###########
            # REPLAY #
            ##########

            # retrieve resources given as arguments
            # NOTE: some resources and actions may depend on others.
            #       Please check the documentation to learn more about this
            resources = args.resources
            actions = args.actions

            # if not specified, replay every resources and actions available with Reven
            # NOTE: some resources and actions cannot be replayed if the snapshot is not prepared
            #       Please check the documentation to learn more about this
            if resources is None:
                resources = [resource["name"] for resource in pm.get_resources_list() if not resource["deprecated"]]
            if actions is None:
                actions = [action["name"] for action in pm.get_actions_list()]

            # start the replay of the scenario
            # resources and actions to replay are required
            # extra optional arguments like `max_icount` or `custom_options` can be specified
            pm.replay_scenario(scenario["id"], resources=resources, actions=actions)
            print("Replay started for scenario %d" % scenario["id"])

            # retrieve the scenario with replay status in it
            scenario = pm.get_scenario(scenario["id"])
            display_resources_actions_progress(scenario["replay"]["resources"], scenario["replay"]["actions"])

            # continuously check the status of scenarios replay until all replays are over
            while not is_replay_finished(scenario):
                sleep(1)

                # retrieve the scenario with replay status in it
                scenario = pm.get_scenario(scenario["id"])
                display_resources_actions_progress(scenario["replay"]["resources"], scenario["replay"]["actions"])

            if scenario["replay"]["status"] == "SUCCESS":
                print("Scenario is replayed")
            else:
                # retrieve logs from task(s) that failed
                display_failed_resources_actions_log(
                    pm, scenario["replay"]["resources"], scenario["replay"]["actions"]
                )

                raise RuntimeError("Scenario failed to replay completely")

        except:  # noqa: E722
            if args.clean_on_failure:
                # if the something failed (scenario creation, record or replay),
                # cleanup created scenario
                if scenario is not None:
                    response = pm.delete_scenario(scenario["id"])
                    print(response["message"])

                # cleanup uploaded binary file
                if binary_file is not None:
                    response = pm.delete_file(binary_file["id"])
                    print(response["message"])

            raise

        print()

        ##############
        # END REPLAY #
        ##############

        # start a Reven server on a specific scenario
        def start_reven_session(scenario_id):
            response = pm.start_reven_session(scenario_id)
            print(response["message"])
            reven_session = response["session"]

            if reven_session["port"] is not None:
                return reven_session

            try:
                for i in range(10):
                    reven_session = pm.get_session(reven_session["id"])

                    if reven_session["port"] is not None:
                        return reven_session

                    if reven_session["status"] != "STARTED":
                        raise RuntimeError("Reven server started but was terminated early")

                    sleep(1)

                raise RuntimeError("Timeout while waiting for the Reven server port")

            except:  # noqa: E722
                # stop the Reven server (session)
                response = pm.stop_session(reven_session["id"])
                print(response["message"])

                raise

        try:
            reven_session = start_reven_session(scenario["id"])
            reven_port = reven_session["port"]

            try:
                # connect to the reven server
                host = urlparse(args.url).hostname
                reven_server = reven2.RevenServer(host, reven_port)

                # TODO: Insert your python code for analysis here
                print("Analysis")

            finally:
                # stop the Reven server (session)
                response = pm.stop_session(reven_session["id"])
                print(response["message"])

        finally:
            # if the export option has been specified, export the generated scenario
            if args.export:
                # you can choose what to export
                # at least the record is exported
                # but you can specify other parameters like replay, light_ossi, light_pdb and user_data
                # recommanded parameters are enabled by default
                export_task = pm.export_scenario(scenario["id"])["task"]
                print("Exporting scenario %d" % scenario["id"])

                # check if the export task is finished
                while not export_task["finished"]:
                    sleep(1)
                    export_task = pm.get_task(export_task["id"])

                    if export_task["status"] == "FAILURE":
                        raise RuntimeError("Export scenario task failed")

                    assert pm.get_scenario(scenario["id"])["archive_status"] != "DOES_NOT_EXIST"

                # retrieve the archive name from the scenario
                archive_path = pm.get_scenario(scenario["id"])["archive_path"]
                print("Scenario %d has been exported. Archive location: %s" % (scenario["id"], archive_path))

                # when the scenario is fully exported (task finished), delete it.
                response = pm.delete_scenario(scenario["id"])
                print(response["message"])

                # you can now import the scenario with the achive name using the `import_scenario` method
                # To retrieve the archive name, get the file name from the path (retrieve from the scenario)
                # >>> import_task = pm.import_scenario(os.path.basename(archive_path))
                # check when the task is over, like above

    except project_manager.ProjectManagerError as pme:
        print(pme)
        exit(1)
    except Exception as e:
        print("Unexpected error: %s" % str(e))
        exit(1)
