"""
This script is an example of automatic multiple records.
It covers the main steps from the scenario creation to the analysis with a Reven server
The multiple record feature is used to record multiple "slices" of a same VM.
For example, continuously recording a VM until a certain event happen can be useful
to record non deterministic scenario,
such as malware that executes code randomly.
Slicing allows the user to only keep the smaller trace containing relevant events.
"""

import argparse
import socket
import sys
from pathlib import Path
from time import sleep
from urllib.parse import urlparse


import reven2
import reven2.preview.project_manager as project_manager

from utils.project_manager import (
    display_failed_resources_actions_log,
    display_resources_actions_progress,
    is_replay_finished,
)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Fully automated scenario creation")

    parser.add_argument("--url", required=True, help="The URL of the Reven project manager")
    parser.add_argument("--snapshot", type=int, required=True, help="The id of the snapshot to use for the scenario")
    parser.add_argument("--live-snp", default=None, help="Name of the live snapshot to use")
    parser.add_argument("--name", default="automatic-scenario", help="The name of the scenario to create")
    parser.add_argument("--description", default="", help="Description of the scenario")
    parser.add_argument("--record-number", type=int, default="5", help="Number of record to do")
    parser.add_argument(
        "--resources", default=None, nargs="+", help="Specific resources you want to replay. By default replay all"
    )
    parser.add_argument(
        "--actions", default=None, nargs="+", help="Specific resources you want to replay. By default replay all"
    )
    parser.add_argument(
        "--clean-on-failure", dest="clean_on_failure", action="store_true", help="Clean created resources upon failure"
    )
    parser.add_argument(
        "--export", dest="export", action="store_true", help="Export and delete scenarios at the end of the script"
    )

    args = parser.parse_args()

    try:
        # connect to the Reven project manager server
        pm = project_manager.ProjectManager(args.url)

        # check that we are using a QEMU snapshot because automatic record is **not supported** for Vbox scenario
        if pm.get_snapshot(args.snapshot)["type"] != "QEMU":
            print("Automatic and multiple record is only available with a QEMU snapshot")
            exit(1)

        ######################
        # SCENARIOS CREATION #
        ######################

        # list to save generated scenarios
        scenarios = []
        # create a new scenario here
        # create scenario with a name and a snapshot (required) and a description (optional)
        # extra optional arguments can be specified like `cdrom_mount_point` or `max_icount`
        try:
            for counter in range(args.record_number):
                scenario = pm.create_scenario(
                    args.name + ("_%02d" % counter),
                    args.snapshot,
                    description="Record " + str(counter) + ": " + args.description,
                )
                # add the scenario to the list
                # Later, a record will be committed into each scenario of the list
                scenarios.append(scenario["id"])
                print('Scenario "%s" created with id %d' % (scenario["name"], scenario["id"]))

            print()

            ##########################
            # END SCENARIOS CREATION #
            ##########################

            ##########
            # RECORD #
            ##########

            # First, start the snapshot
            # a specific (optional) QEMU live snapshot can be started (if specified like here)
            # extra optional arguments can be specified, like `network`, `ram_size`, `custom_options`
            response = pm.start_qemu_snapshot_session(args.snapshot, live_snapshot=args.live_snp)
            print(response["message"])
            socket_path = Path(response["session"]["socket_path"])
            precise_slicing = False
            if socket_path.exists():
                print("QEMU socket is accessible, activating precise slicing")
                precise_slicing = True
            else:
                print("QEMU socket is not accessible, precise slicing is not possible")
                print("Expected socket: %s" % socket_path)
                print("To easily enable this feature, simply run this script on the same machine that runs Quasar.")

            def pause_vm():
                if precise_slicing:
                    session_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                    session_socket.connect(str(socket_path))
                    session_socket.send("stop\n".encode("utf-8"))
                    session_socket.close()
                    print("VM paused... ", end="")
                    sys.stdout.flush()

            def resume_vm():
                if precise_slicing:
                    session_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                    session_socket.connect(str(socket_path))
                    session_socket.send("cont\n".encode("utf-8"))
                    session_socket.close()
                    print("VM resumed... ", end="")
                    sys.stdout.flush()

            # we need to keep the VM session to start/stop/commit records and stop the session later
            session_id = response["session"]["id"]
            try:
                for scenario_id in scenarios:
                    pm.start_record(session_id)
                    # wait for 5 seconds for each record (just a record where nothing happen in the VM)
                    # need the VM introspection/instrumentation to do something useful with it
                    print("Recording... ", end="")
                    sys.stdout.flush()
                    resume_vm()
                    sleep(5)
                    pause_vm()
                    # stop the record
                    response = pm.stop_record(session_id)
                    print("OK")
                    # retrieve the record name to be able to commit it into a scenario
                    record_name = response["record_name"]
                    # commit the record in a scenario
                    response = pm.commit_record(session_id, scenario_id, record=record_name)
                    print(response["message"])
            finally:
                # stop the session when we are done recording scenarios
                # NOTE: you will lose uncommitted records upon stopping the session
                response = pm.stop_session(session_id)
                print(response["message"])

            print()

            ##############
            # END RECORD #
            ##############

            ###########
            # REPLAY #
            ##########

            # retrieve resources given as arguments
            # NOTE: some resources and actions may depend on others.
            #       Please check the documentation to learn more about this
            resources = args.resources
            actions = args.actions

            # if not specified, replay every resources and actions available with Reven
            # NOTE: some resources and actions cannot be replayed if the snapshot is not prepared
            #       Please check the documentation to learn more about this
            if resources is None:
                resources = [resource["name"] for resource in pm.get_resources_list() if not resource["deprecated"]]
            if actions is None:
                actions = [action["name"] for action in pm.get_actions_list()]

            if args.export:
                # copy the scenario list, will be needed to export them all
                # the current `scenarios` list will be modified
                created_scenarios = scenarios.copy()

            for scenario_id in scenarios:
                # start the replay on a specific scenario
                # resources and actions to replay are required
                # extra optional arguments like `max_icount` or `custom_options` can be specified
                pm.replay_scenario(scenario_id, resources=resources, actions=actions)
                print("Replay started for scenario %d" % scenario_id)
                # retrieve scenario with resources and actions in it
                scenario = pm.get_scenario(scenario_id)
                display_resources_actions_progress(scenario["replay"]["resources"], scenario["replay"]["actions"])
            # continuously check the status of scenarios replay until all replays are over
            while scenarios:
                for scenario_id in scenarios:
                    print("Scenario %d" % scenario_id)
                    scenario = pm.get_scenario(scenario_id)
                    display_resources_actions_progress(scenario["replay"]["resources"], scenario["replay"]["actions"])
                    # check if the replay is finished
                    if is_replay_finished(scenario):
                        if scenario["replay"]["status"] != "SUCCESS":
                            # retrieve logs from task(s) that failed
                            display_failed_resources_actions_log(
                                pm, scenario["replay"]["resources"], scenario["replay"]["actions"]
                            )

                            raise RuntimeError("Scenario %d failed to replay completly" % scenario_id)

                        print("Replay of scenario %d is completed" % scenario_id)
                        # remove the scenario with completed replay of the `scenarios` list
                        scenarios[:] = list(filter(lambda x: x != scenario_id, scenarios))
                    sleep(1)
                    print()
            print("All scenarios are replayed")

        except:  # noqa: E722
            if args.clean_on_failure:
                # if the something failed (scenario creation, record or replay),
                # cleanup created scenarios
                for scenario_id in scenarios:
                    response = pm.delete_scenario(scenario_id)
                    print(response["message"])

            raise

        print()

        ##############
        # END REPLAY #
        ##############

        # start a Reven server on a specific scenario
        def start_reven_session(scenario_id):
            response = pm.start_reven_session(scenario_id)
            print(response["message"])
            reven_session = response["session"]

            if reven_session["port"] is not None:
                return reven_session

            try:
                for i in range(10):
                    reven_session = pm.get_session(reven_session["id"])

                    if reven_session["port"] is not None:
                        return reven_session

                    if reven_session["status"] != "STARTED":
                        raise RuntimeError("Reven server started but was terminated early")

                    sleep(1)

                raise RuntimeError("Timeout while waiting for the Reven server port")

            except:  # noqa: E722
                # stop the Reven server (session)
                response = pm.stop_session(reven_session["id"])
                print(response["message"])

                raise

        try:
            reven_session = start_reven_session(scenario_id)
            reven_port = reven_session["port"]

            try:
                # connect to the reven server
                host = urlparse(args.url).hostname
                reven_server = reven2.RevenServer(host, reven_port)

                # TODO: Insert your python code for analysis here
                print("Analysis")

            finally:
                # stop the Reven server (session)
                response = pm.stop_session(reven_session["id"])
                print(response["message"])

        finally:
            # if the export option has been specified, export all generated scenario
            if args.export:
                for scenario_id in created_scenarios:
                    # you can choose what to export
                    # at least the record is exported
                    # but you can specify other parameters like replay, light_ossi, light_pdb and user_data
                    # recommanded parameters are enabled by default
                    export_task = pm.export_scenario(scenario_id)["task"]
                    print("Exporting scenario %d" % scenario_id)

                    # check if the export task is finished
                    while not export_task["finished"]:
                        sleep(1)
                        export_task = pm.get_task(export_task["id"])

                        if export_task["status"] == "FAILURE":
                            raise RuntimeError("Export scenario task failed")

                        assert pm.get_scenario(scenario_id)["archive_status"] != "DOES_NOT_EXIST"

                    # retrieve the archive name from the scenario
                    archive_path = pm.get_scenario(scenario_id)["archive_path"]
                    print("Scenario %d has been exported. Archive location: %s" % (scenario_id, archive_path))

                    # when the scenario is fully exported (task finished), delete it.
                    response = pm.delete_scenario(scenario_id)
                    print(response["message"])

                    # you can now import the scenario with the achive name using the `import_scenario` method
                    # To retrieve the archive name, get the file name from the path (retrieve from the scenario)
                    # >>> import_task = pm.import_scenario(os.path.basename(archive_path))
                    # check when the task is over, like above

    except project_manager.ProjectManagerError as pme:
        print(pme)
        exit(1)
    except Exception as e:
        print("Unexpected error: %s" % str(e))
        exit(1)
